(* This file provides functions for translating the Kaleidoscope AST into LLVM IR *)

(* Standard libraries *)
open Array
open List
open Printf

(* LLVM related libraries *)
open Llvm

(* This project's libraries *)
open KaleidoscopeAST

exception Error of string

let working_module = create_module "Kaleidoscope JIT"
let builder = builder ()
let named_values:(string, llvalue) Hashtbl.t = Hashtbl.create 10

let rec translate_function node =
	let Function(p, e) = node in
		Hashtbl.clear named_values;
		let f = translate_prototype p in
			(* Create a new basic block to start insertion into. *)
			let bb = append_block "entry" f in
				position_at_end bb builder;
				
				try
					let ret_val = translate_expression e in
						(* Finish off the function. *)
						let _ = build_ret ret_val builder in
							(* Validate the generated code, checking for consistency. *)
							Llvm_analysis.assert_valid_function f;

							f
				with error ->
					delete_function f;
					raise error

and translate_prototype node =
	let Prototype(name, args) = node in
		(* Make the function type: double(double,double) etc. *)
		let doubles = Array.make (List.length args) double_type in
			let ft = function_type double_type doubles in
				let f =
					match lookup_function name working_module with
						  None ->
							declare_function name ft working_module
						(* If 'f' conflicted, there was already something named 'name'. If it
					  	 * has a body, don't allow redefinition or reextern. *)
						| Some(f) ->
							(* If 'f' already has a body, reject this. *)
							if block_begin f <> At_end f then
								raise (Error"Redefinition of function");

							(* If 'f' took a different number of arguments, reject. *)
							if element_type (type_of f) <> ft then
								raise (Error"Redefinition of function with different # args.");
							f
				in
					(* Set names for all arguments. *)
					let name_args i a =
						let n = nth args i in
							set_value_name n a;
							Hashtbl.add named_values n a;
					in
					
					Array.iteri name_args (params f);
					f

and translate_expression node =
	match node with
		  Variable(name) ->
		  	begin
		  		try
		  			Hashtbl.find named_values name
		  		with
		  			Not_found ->
		  				raise (Error (Printf.sprintf "Uninitialized variable '%s.'" name))
		  	end
		| Number(num) ->
			const_float double_type num
		| Binary(op, lhs, rhs) ->
			let lhs_value = translate_expression lhs in
				let rhs_value = translate_expression rhs in
					begin
						match op with
							  Plus ->
							  	build_add lhs_value rhs_value "addtmp" builder
							| Sub ->
								build_sub lhs_value rhs_value "subtmp" builder
							| Mul ->
								build_mul lhs_value rhs_value "multmp" builder
							| Div ->
								build_fdiv lhs_value rhs_value "divtmp" builder
							| Lt ->
								let i = build_fcmp Fcmp.Ult lhs_value rhs_value "cmptmp" builder in
									build_uitofp i double_type "booltmp" builder
					end
		| Call(name, args) ->
			let callee =
				match lookup_function name working_module with
					  Some(callee) -> callee
					| None ->
						raise (Error "Unknown function referenced.")
				in
					let params = params callee in
						(* If argument mismatch error. *)
						if Array.length params == List.length args then
							()
						else
							raise (Error "Incorrect # arguments passed.");
						
						let args = Array.of_list (List.map translate_expression args) in
							build_call callee args "calltmp" builder

