(* This file provides functions for translating the Kaleidoscope AST into LLVM IR *)

(* Standard libraries *)
open Array
open List
open Printf

(* LLVM related libraries *)
open Llvm

(* This project's libraries *)
open KaleidoscopeAST

exception Error of string

let working_module = create_module "Kaleidoscope JIT"
let builder = builder ()
let named_values:(string, llvalue) Hashtbl.t = Hashtbl.create 10

(* Create an alloca instruction in the entry block of the function. This is  *)
(* is used for mutable variables etc.                                        *)
let create_entry_block_alloca f var_name =
	let builder = builder_at (instr_begin (entry_block f)) in
		build_alloca double_type var_name builder

(* Create an alloca for each argument and register the argument in the       *)
(* symbol table so that references to it will succeed.                       *)
let create_argument_allocas f proto =
	let Prototype(_,args) = proto in
		let iter_fun i ai =
			let var_name = nth args i in
				(* Create an alloca for this variable. *)
				let alloca = create_entry_block_alloca f var_name in
					ignore(build_store ai alloca builder);
					Hashtbl.add named_values var_name alloca;
		in
		
		Array.iteri iter_fun (params f)

let rec translate_function node =
	let Function(p, e) = node in
		Hashtbl.clear named_values;
		let f = translate_prototype p in
			(* Create a new basic block to start insertion into. *)
			let bb = append_block "entry" f in
				position_at_end bb builder;
				
				try
					(* Add all arguments to the symbol table and create *)
					(*  their allocas.                                  *)
					create_argument_allocas f p;
					
					let ret_val = translate_expression e in
						(* Finish off the function. *)
						let _ = build_ret ret_val builder in
							(* Validate the generated code, checking for consistency. *)
							Llvm_analysis.assert_valid_function f;

							f
				with error ->
					delete_function f;
					raise error

and translate_prototype node =
	let Prototype(name, args) = node in
		(* Make the function type: double(double,double) etc. *)
		let doubles = Array.make (List.length args) double_type in
			let ft = function_type double_type doubles in
				let f =
					match lookup_function name working_module with
						  None ->
							declare_function name ft working_module
						(* If 'f' conflicted, there was already something named 'name'. If it
					  	 * has a body, don't allow redefinition or reextern. *)
						| Some(f) ->
							(* If 'f' already has a body, reject this. *)
							if block_begin f <> At_end f then
								raise (Error"Redefinition of function");

							(* If 'f' took a different number of arguments, reject. *)
							if element_type (type_of f) <> ft then
								raise (Error"Redefinition of function with different # args.");
							f
				in
					(* Set names for all arguments. *)
					let name_args i a =
						let n = nth args i in
							set_value_name n a;
							Hashtbl.add named_values n a;
					in
					
					Array.iteri name_args (params f);
					f

and translate_expression node =
	match node with
		  Variable(name) ->
		  	begin
		  		let v =
		  			try
			  			Hashtbl.find named_values name
			  		with Not_found ->
		  				raise (Error (Printf.sprintf "Uninitialized variable '%s'." name))
		  		in
		  		build_load v name builder
		  	end
		| Number(num) ->
			const_float double_type num
		| Binary(op, lhs, rhs) ->
			begin
				let rhs_value = translate_expression rhs in
					match op with
						  Assign ->
						  	(* Special case '=' because we don't want to   *)
							(* emit the LHS as an expression.              *)
							let name =
								match lhs with
								  Variable name -> name
								| _ -> raise (Error "destination of '=' must be a variable")
							in

							(* Lookup the name. *)
							let variable =
								try
									Hashtbl.find named_values name
								with
									Not_found -> raise (Error "unknown variable name")
							in
					
							ignore(build_store rhs_value variable builder);
							rhs_value
						| Plus ->
							build_add (translate_expression lhs) rhs_value "addtmp" builder
						| Sub ->
							build_sub (translate_expression lhs) rhs_value "subtmp" builder
						| Mul ->
							build_mul (translate_expression lhs) rhs_value "multmp" builder
						| Div ->
							build_fdiv (translate_expression lhs) rhs_value "divtmp" builder
						| Lt ->
							let lhs_value = translate_expression lhs in
								let i = build_fcmp Fcmp.Ult lhs_value rhs_value "cmptmp" builder in
									build_uitofp i double_type "booltmp" builder
			end
		| Call(name, args) ->
			let callee =
				match lookup_function name working_module with
					  Some(callee) -> callee
					| None ->
						raise (Error "Unknown function referenced.")
				in
					let params = params callee in
						(* If argument mismatch error. *)
						if Array.length params == List.length args then
							()
						else
							raise (Error "Incorrect # arguments passed.");
						
						let args = Array.of_list (List.map translate_expression args) in
							build_call callee args "calltmp" builder
		| VarDef(defs, e) ->
			let old_bindings = ref [] in
			let f = block_parent (insertion_block builder) in
			
			(* Emit the initializer before adding the variable to scope,  *)
			(* this prevents the initializer from referencing the         *)
			(* variable itself, and permits stuff like this:              *)
			(*   var a = 1 in                                             *)
			(*     var a = a in ...   # refers to outer 'a'.              *)
			let install_var (var_name, init) =
				let alloca = create_entry_block_alloca f var_name in
					let init_value = translate_expression init in
					ignore(build_store init_value alloca builder);
					
					(* Remember the old variable binding so that we can *)
					(* restore the binding when we unrecurse.           *)
					begin
						try
							let old_value = Hashtbl.find named_values var_name in
							old_bindings := (var_name, old_value) :: !old_bindings;
						with Not_found -> ()
					end;
					
					Hashtbl.add named_values var_name alloca;
			in
			
			(* Register all variables and emit their initializer. *)
			List.iter install_var defs;
			
			(* Translate the body now that all of the variables are in    *)
			(* scope.                                                     *)
			let body_value = translate_expression e in
			
			(* Pop all our variables from scope. *)
			List.iter (fun (var_name, old_value) -> Hashtbl.add named_values var_name old_value) !old_bindings;
			
			(* Return the value produced by evaluating the body. *)
			body_value
		| If(cond, then_, else_) ->
			let cond = translate_expression cond in
			(* Convert condition to a bool by comparing equal to 0.0 *)
			let zero = const_float double_type 0.0 in
			let cond_value = build_fcmp Fcmp.One cond zero "ifcond" builder in
			
			(* Grab the first block so that we might later add the        *)
			(* conditional branch to it at the end of the function.       *)
			let start_bb = insertion_block builder in
			let f = block_parent start_bb in
								
				let then_bb = append_block "then" f in
					position_at_end then_bb builder;
									
					let then_value = translate_expression then_ in
					(* Codegen of 'then' can change the current block,  *)
					(* update then_bb for the phi. We create a new name *)
					(* because one is used for the phi node, and the    *)
					(* other is used for the conditional branch.  This  *)
					(* call grabs whatever block the builder is         *)
					(* set to insert into.                              *)
					let new_then_bb = insertion_block builder in
					
				let else_bb = append_block "else" f in
					position_at_end else_bb builder;
					let else_value = translate_expression else_ in
					(* Codegen of 'else' can change the current block,  *)
					(* update else_bb for the phi. *)
					let new_else_bb = insertion_block builder in
				
				(* Emit merge block. *)
				let merge_bb = append_block "ifcont" f in
					position_at_end merge_bb builder;
					let incoming = [(then_value, new_then_bb); (else_value, new_else_bb)] in
					let phi = build_phi incoming "iftmp" builder in
				
				(* Return to the start block to add the conditional      *)
				(* branch.                                               *)
				position_at_end start_bb builder;
				ignore (build_cond_br cond_value then_bb else_bb builder);
				
				(* Set a unconditional branch at the end of the 'then'   *)
				(* block and the 'else' block to the 'merge' block.      *)
				position_at_end new_then_bb builder;
				ignore (build_br merge_bb builder);
				
				position_at_end new_else_bb builder;
				ignore (build_br merge_bb builder);
				
				(* Finally, set the builder to the end of the merge      *)
				(* block.                                                *)
				position_at_end merge_bb builder;
				
				phi
		| For(var_name, start, end_, step, body) ->
			let f = block_parent (insertion_block builder) in
			
			(* Create an alloca for the variable in the entry block.      *)
			let alloca = create_entry_block_alloca f var_name in
			
			(* Emit the start code first, without 'variable' in scope.    *)
			let start_val = translate_expression start in
			
			(* Store the value into the alloca. *)
			ignore(build_store start_val alloca builder);
			
			(* Make the new basic block for the loop header, inserting    *)
			(* after current block. *)
			let loop_bb = append_block "loop" f in
			
			(* Insert an explicit fall through from the current block to  *)
			(* the loop_bb.                                               *)
			ignore (build_br loop_bb builder);
			
			(* Start insertion in loop_bb. *)
			position_at_end loop_bb builder;
			
			(* Within the loop, the variable is defined equal to the PHI  *)
			(* node. If it shadows an existing variable, we have to       *)
			(* restore it, so save it now. *)
			let old_val =
				try Some (Hashtbl.find named_values var_name) with Not_found -> None
			in
			Hashtbl.add named_values var_name alloca;
			
			(* Emit the body of the loop.  This, like any other expr, can *)
			(* change the current BB.  Note that we ignore the value      *)
			(* computed by the body, but don't allow an error.            *)
			ignore (translate_expression body);
			
			(* Compute the end condition. *)
			let end_cond = translate_expression end_ in
			
			(* Reload, increment, and restore the alloca. This handlesthe *)
			(* case where the body of the loop mutates the variable.      *)
			let cur_var = build_load alloca var_name builder in
			let next_var = build_add cur_var (translate_expression step) "nextvar" builder in
			ignore(build_store next_var alloca builder);
			
			(* Convert condition to a bool by comparing equal to 0.0. *)
			let zero = const_float double_type 0.0 in
			let end_cond = build_fcmp Fcmp.One end_cond zero "loopcond" builder in
			
			let after_bb = append_block "afterloop" f in
				(* Insert the conditional branch into the end of loop_end_bb. *)
				ignore (build_cond_br end_cond loop_bb after_bb builder);

				(* Any new code will be inserted in after_bb. *)
				position_at_end after_bb builder;

				(* Restore the unshadowed variable. *)
				begin
					match old_val with
						  Some old_val -> Hashtbl.add named_values var_name old_val
						| None -> ()
				end;

				(* for expr always returns 0.0. *)
				const_null double_type

